module Exam.Exercise2.All2 where

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_; refl; sym; cong; subst)
open import Relation.Nullary using (Dec; yes; no; ¬_)
open import Data.Maybe using (Maybe; just; nothing)
open import Data.Nat as Nat using (ℕ; zero; suc)
open import Data.Empty using (⊥; ⊥-elim)
open import Data.List using (List; _∷_; []; _++_; [_])

open import Exam.Exercise2.Sublist

-- The All₂ type is similar to the All type you've met in the chapter about Lists.
--
-- Here is a quick repetition of the All type:
--   If we have a unary relation on some type A, e.g.
--     R : A → Set
--   then the All type gives us a unary relation on lists of A for free:
--     All R : List A → Set
--   such that
--     All R (1 ∷ 2 ∷ 3 ∷ [])
--   is true iff R is true for all elements of the list, i.e.
--     R 1 × R 2 × R 3
--   This illustrates, that a proof for (All R (1 ∷ 2 ∷ 3 ∷ [])) is basically
--   a list of proofs that R holds for the individual elements 1, 2, and 3.
--   This intuition, is also visible in the definition of the All type,
--   where the constructors are named like the list constructors [] and _∷_.
--   Hence, if we have proofs
--     p₁ : R 1
--     p₂ : R 2
--     p₃ : R 3
--   we can construct
--    p₁ ∷ p₂ ∷ p₃ ∷ []  :  All R (1 ∷ 2 ∷ 3 ∷ [])
--
-- The All₂ type behaves just like the All type, but for binary relations
-- instead of unary relations. If we have a binary relation
--   R : A → B → Set
-- then the All₂ type gives us a binary relation on lists
--   All₂ R : List A → List B → Set
-- such that
--   All R (1 ∷ 2 ∷ 3 ∷ []) (4 ∷ 5 ∷ 6 ∷ [])
-- is true iff R is true for the "zipping" of the elements of both lists, i.e.
--   R 1 4 × R 2 5 × R 3 6
--
-- The All₂ is used in SoundnessSmallStep.agda to define the Σ ⊢ σ relation,
-- which relates a list of values σ with a list of types Σ using the
-- typing relation for R.

data All₂ {A B : Set} (R : A → B → Set) : List A → List B → Set where
  [] : All₂ R [] []
  _∷_ : ∀ {xs ys x y} → R x y → All₂ R xs ys → All₂ R (x ∷ xs) (y ∷ ys)

-- If two lists are related by an All₂, then they must have equal length.
postulate
  all₂-length : ∀ {A B : Set} {R : A → B → Set} {xs : List A} {ys : List B} →
    All₂ R xs ys →
    length xs ≡ length ys

-- Recall, how the All type basically corresponds to a list of proofs.
-- For lists, we have a lookup operation that gives us the element at a certain index.
-- This lemma allows us to do the same for the All₂ type:
-- if x and y are the elements at index n, then we can lookup the proof R x y.
postulate
  lookup-all : ∀ {A B : Set} {R : A → B → Set} {xs : List A} {ys : List B} {n : ℕ} {x : A} {y : B} →
    All₂ R xs ys →
    lookup n xs ≡ just x →
    lookup n ys ≡ just y →
    R x y

-- Just like the lookup-all corresponds to lookup, update-all corresponds to update:
-- If we know that (All₂ R xs ys) is true, then update-all allows us to replace
-- a single proof in the "proof list" with a proof that refers to a different y.
-- Hence, the All₂ returned by lookup-all, also refers to an updated list of ys:
--   All₂ R xs (update n y ys)
postulate
  update-all : ∀ {A B : Set} {R : A → B → Set} {xs : List A} {ys : List B} {n x y} →
    All₂ R xs ys →
    lookup n xs ≡ just x →
    R x y →
    All₂ R xs (update n y ys)
