module Exam.Exercise2.SemanticsSmallStep where

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_; refl; sym)
open import Relation.Nullary                      using (Dec; yes; no; ¬_)
open import Data.String                           using (_≟_)
open import Data.List                             using (List; _∷_; [])
open import Data.Nat as Nat                       using (ℕ; zero; suc)
open import Data.Maybe                            using (Maybe; just; nothing)

open import Exam.Exercise2.Syntax
open import Exam.Exercise2.Substitution
open import Exam.Exercise2.Sublist

-- Small-Step Semantics --------------------------------------------------------

data Value : Term → Set where

  V-ƛ :
    Value (ƛ x ⇒ M)

  V-zero :
    Value `zero

  V-suc :
    Value M →
    ----------
    Value (`suc M)

  -- We consider two new terms as values:

  V-loc :
    Value (`loc ℓ)

  V-tt :
    Value `tt

-- As described in the pdf, we describe the store as a list of values,
-- where the value at index ℓ describes the value at location ℓ.

Store = List Term

variable
  σ σ′ σ′′ σ₁ σ₂ : Store

-- Next, we define the small-step semantics:

infix 4 ⟨_,_⟩→⟨_,_⟩

data ⟨_,_⟩→⟨_,_⟩ : Store → Term → Store → Term → Set where

  -- For the old terms, the semantics stays the same. We simply leave the store
  -- σ unmodified, since the old terms have nothing to do with memory.

  β-ƛ :

    Value N →
    ------------------------------
    ⟨ σ , (ƛ x ⇒ M) · N ⟩→⟨ σ , M [ x := N ] ⟩

  β-zero :

    ⟨ σ , case `zero [zero⇒ M |suc x ⇒ N ] ⟩→⟨ σ , M ⟩

  β-suc :

    Value L →
    --------------------------------------------------
    ⟨ σ , case (`suc L) [zero⇒ M |suc x ⇒ N ] ⟩→⟨ σ , N [ x := L ] ⟩

  β-μ :

    ⟨ σ , μ x ⇒ M ⟩→⟨ σ , M [ x := μ x ⇒ M ] ⟩

  -- For the new terms, the store is important:

  β-new :                                       -- The (`new M) term extends the
    Value M →                                   -- store σ with a new memory cell
    ------------------------------------------  -- containing M, and returns the
    ⟨ σ , `new M ⟩→⟨ M ∷ σ , `loc (length σ) ⟩  -- location of this cell

  β-! :                                         -- The (! (`loc ℓ)) term reduces
    lookup ℓ σ ≡ just M →                       -- to the value stored at
    ----------------------------                -- location ℓ in σ.
    ⟨ σ , ! (`loc ℓ) ⟩→⟨ σ , M ⟩                -- The store σ is left unchanged.

  β-≔ :                                         -- The (`loc ℓ ≔ N) term, updates
    Value N →                                   -- the store at location ℓ with
    ----------------------------------------    -- a new value N, and reduces
    ⟨ σ , `loc ℓ ≔ N ⟩→⟨ update ℓ N σ , `tt ⟩   -- to the unit value `tt.

  -- The ξ-rules work the same as always, in the sense that they only redirect
  -- computation to subterms. Since those subterms can contain terms
  -- related to references, evaluating the subterms may modify the store.
  -- Hence, both the new and old ξ-rules allow the evaluation of the subterms,
  -- to modify the store from σ to σ´.

  ξ-·₁ :

    ⟨ σ , M     ⟩→⟨ σ′ , M′     ⟩ →
    --------------------
    ⟨ σ , M · N ⟩→⟨ σ′ , M′ · N ⟩

  ξ-·₂ :

    Value M →
    ⟨ σ ,     N ⟩→⟨ σ′ ,     N′ ⟩ →
    --------------------
    ⟨ σ , M · N ⟩→⟨ σ′ , M · N′ ⟩

  ξ-suc :

    ⟨ σ ,      M ⟩→⟨ σ′ ,      M′ ⟩ →
    --------------------
    ⟨ σ , `suc M ⟩→⟨ σ′ , `suc M′ ⟩

  ξ-case : ∀ {σ σ′} →

    ⟨ σ ,      L                       ⟩→⟨ σ′ ,      L′                       ⟩ →
    ---------------------------------------------------------------------------
    ⟨ σ , case L [zero⇒ M |suc x ⇒ N ] ⟩→⟨ σ′ , case L′ [zero⇒ M |suc x ⇒ N ] ⟩

  ξ-new :
    ⟨ σ ,      M ⟩→⟨ σ′ ,      M′ ⟩ →
    -------------------------------
    ⟨ σ , `new M ⟩→⟨ σ′ , `new M′ ⟩

  ξ-! :
    ⟨ σ ,   M ⟩→⟨ σ′ ,   M′ ⟩ →
    -------------------------
    ⟨ σ , ! M ⟩→⟨ σ′ , ! M′ ⟩

  ξ-≔₁ :
    ⟨ σ , M     ⟩→⟨ σ′ , M′     ⟩ →
    -----------------------------
    ⟨ σ , M ≔ N ⟩→⟨ σ′ , M′ ≔ N ⟩

  ξ-≔₂ :
    ⟨ σ ,     N ⟩→⟨ σ′ ,     N′ ⟩ →
    -----------------------------
    ⟨ σ , M ≔ N ⟩→⟨ σ′ , M ≔ N′ ⟩

-- Extending the reflexive, transitive closure of the small-step semantics,
-- to incorporate stores is straightforward:

data ⟨_,_⟩→*⟨_,_⟩ : Store → Term → Store → Term → Set where

  —↠-refl :
    ⟨ σ , M ⟩→*⟨ σ , M ⟩

  —↠-step :
    ⟨ σ , M ⟩→⟨ σ′ , N ⟩ →
    ---------------------
    ⟨ σ , M ⟩→*⟨ σ′ , N ⟩

  —↠-trans :
    ⟨ σ  , M ⟩→*⟨ σ′  , N ⟩ →
    ⟨ σ′ , N ⟩→*⟨ σ′′ , L ⟩ →
    ----------
    ⟨ σ  , M ⟩→*⟨ σ′′ , L ⟩
