module Exam.Exercise2.Sublist where

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_; refl; sym; cong)
open import Relation.Nullary using (Dec; yes; no; ¬_)
open import Data.Maybe using (Maybe; just; nothing)
open import Data.Nat as Nat using (ℕ; zero; suc; _≤_; s≤s; z≤n)
open import Data.Empty using (⊥; ⊥-elim)
open import Data.Product
  using (_×_; proj₁; proj₂; ∃; ∃-syntax)
  renaming (_,_ to ⟨_,_⟩)
open import Data.List using (List; _∷_; []; _++_; [_])

-- In this file, we define certain list operations and prove lemmas about them.
--
-- Recall, that our language models a store as a list of values, where the value
-- at index ℓ corresponds to the value at memory location ℓ.
-- The semantics, needs to modify this store by retrieving or updating the
-- values in the list. In this file we provide the necessary definitions and
-- theory for those store modifications.


-- The length of a list.
-- (We don't use the definition from the standard library, as it is implemented
-- using fold, which can lead to confusing looking goals.)
length : ∀ {A : Set} → List A → ℕ
length []       = zero
length (_ ∷ xs) = suc (length xs)

-- Retrieve the list element at a certain index, if it exists.
lookup : {A : Set} → ℕ → List A → Maybe A
lookup n [] = nothing
lookup n (x ∷ xs) with n Nat.≟ length xs
... | yes _ = just x
... | no  _ = lookup n xs

-- Replace the list element at a certain index, if it exists.
update : {A : Set} → ℕ → A → List A → List A
update n x' [] = []
update n x' (x ∷ xs) with n Nat.≟ length xs
... | yes _ = x' ∷ xs
... | no  _ = x ∷ update n x' xs


-- Note, that lookup and update index the list from the *right*:

_ : lookup 0 (1 ∷ 2 ∷ 3 ∷ []) ≡ just 3
_ = refl

_ : update 0 42 (1 ∷ 2 ∷ 3 ∷ []) ≡ 1 ∷ 2 ∷ 42 ∷ []
_ = refl

-- Indexing from the right will be helpful when specifying the semantics
-- of a `new term:
--
--   ⟨ σ , `new `zero ⟩→⟨ `zero ∷ σ , `loc (length σ) ⟩
--
-- Here we extend the current store σ with a new location containing the value `zero.
-- Since we index from the right, we can lookup the new value at index (length σ).
--
-- If we had chosen indexing from the left, then the new value would be at index
-- 0, however the location of all the old values would have changed: what was
-- previously found at index n is now found at index n+1.
--
-- By indexing from the right, but extending the list from the left,
-- all old values remain at the same location.


-- x ≟ x always give us a proof that x is indeed equal to x.
postulate
  ≟-refl : ∀ (x : ℕ) → (x Nat.≟ x) ≡ yes refl

postulate
  lookup-length : ∀ {A : Set} (xs : List A) (x : A) →
    lookup (length xs) (x ∷ xs) ≡ just x


-- The _⊆_ relation expresses that one list is the suffix of another list:
infix 4 _⊆_
_⊆_ : ∀ {A : Set} → List A → List A → Set
xs ⊆ ys = ∃[ zs ] (ys ≡ zs ++ xs)


-- Any list is a suffix of itself
postulate
  ⊆-refl : ∀ {A : Set} {xs : List A} → xs ⊆ xs

-- The suffix-relation is transitive.
-- (You don't need this lemma, but you can prove it for bonus points.)
postulate
  ⊆-trans : ∀ {A : Set} {xs ys zs : List A} → xs ⊆ ys → ys ⊆ zs → xs ⊆ zs

-- The suffix-relation is anti-symmetric.
-- (You don't need this lemma, but you can prove it for bonus points.)
postulate
  ⊆-anti-sym : ∀ {A : Set} {xs ys : List A} → xs ⊆ ys → ys ⊆ xs → xs ≡ ys


-- If we know that xs has an element x at index n,
-- then any (zs ++ xs) also has the element x at index n.
-- (Remember, that lookup counts indices from the right).
postulate
  ⊆-preserves-lookup : ∀ {A : Set} {xs ys : List A} {n x} →
    xs ⊆ ys →
    lookup n xs ≡ just x →
    lookup n ys ≡ just x
